@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright (C) 2010 Radim Rehurek <radimrehurek@seznam.cz>
-# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
+# Licensed under the GNU LGPL v2.1 - https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 
 
 """Module for `Latent Semantic Analysis (aka Latent Semantic Indexing)
@@ -23,7 +23,7 @@ combination of which effectively and transparently allows building LSI models fo
 * distributed computing for very large corpora, making use of a cluster of
   machines
 
-Wall-clock `performance on the English Wikipedia <http://radimrehurek.com/gensim/wiki.html>`_
+Wall-clock `performance on the English Wikipedia <https://radimrehurek.com/gensim/wiki.html>`_
 (2G corpus positions, 3.2M documents, 100K features, 0.5G non-zero entries in the final TF-IDF matrix),
 requesting the top 400 LSI factors:
 
@@ -43,11 +43,13 @@ with dual core Xeon 2.0GHz, 4GB RAM, ATLAS
 
 Examples
 --------
->>> from gensim.test.utils import common_dictionary, common_corpus
->>> from gensim.models import LsiModel
->>>
->>> model = LsiModel(common_corpus, id2word=common_dictionary)
->>> vectorized_corpus = model[common_corpus]  # vectorize input copus in BoW format
+.. sourcecode:: pycon
+
+    >>> from gensim.test.utils import common_dictionary, common_corpus
+    >>> from gensim.models import LsiModel
+    >>>
+    >>> model = LsiModel(common_corpus, id2word=common_dictionary)
+    >>> vectorized_corpus = model[common_corpus]  # vectorize input copus in BoW format
 
 
 .. [1] The stochastic algo could be distributed too, but most time is already spent
@@ -59,16 +61,19 @@ Examples
 
 import logging
 import sys
+import time
 
 import numpy as np
+try:
+    from scipy.sparse import csc_matvecs
+except ImportError:
+    # the function is no longer public in scipy 1.14+
+    from scipy.sparse._sparsetools import csc_matvecs
 import scipy.linalg
-import scipy.sparse
-from scipy.sparse import sparsetools
-from six import iterkeys
-from six.moves import xrange
 
 from gensim import interfaces, matutils, utils
 from gensim.models import basemodel
+from gensim.utils import is_empty
 
 logger = logging.getLogger(__name__)
 
@@ -161,8 +166,11 @@ class Projection(utils.SaveLoad):
     via  :meth:`~gensim.models.lsimodel.Projection.merge`. This is how incremental training actually happens.
 
     """
-    def __init__(self, m, k, docs=None, use_svdlibc=False, power_iters=P2_EXTRA_ITERS,
-                 extra_dims=P2_EXTRA_DIMS, dtype=np.float64):
+
+    def __init__(
+            self, m, k, docs=None, use_svdlibc=False, power_iters=P2_EXTRA_ITERS,
+            extra_dims=P2_EXTRA_DIMS, dtype=np.float64, random_seed=None,
+    ):
         """Construct the (U, S) projection from a corpus.
 
         Parameters
@@ -174,7 +182,7 @@ class Projection(utils.SaveLoad):
         docs : {iterable of list of (int, float), scipy.sparse.csc}
             Corpus in BoW format or as sparse matrix.
         use_svdlibc : bool, optional
-            If True - will use `sparsesvd library <https://pypi.python.org/pypi/sparsesvd/>`_,
+            If True - will use `sparsesvd library <https://pypi.org/project/sparsesvd/>`_,
             otherwise - our own version will be used.
         power_iters: int, optional
             Number of power iteration steps to be used. Tune to improve accuracy.
@@ -182,11 +190,15 @@ class Projection(utils.SaveLoad):
             Extra samples to be used besides the rank `k`. Tune to improve accuracy.
         dtype : numpy.dtype, optional
             Enforces a type for elements of the decomposed matrix.
+        random_seed: {None, int}, optional
+            Random seed used to initialize the pseudo-random number generator,
+            a local instance of numpy.random.RandomState instance.
 
         """
         self.m, self.k = m, k
         self.power_iters = power_iters
         self.extra_dims = extra_dims
+        self.random_seed = random_seed
         if docs is not None:
             # base case decomposition: given a job `docs`, compute its decomposition,
             # *in-core*.
@@ -194,7 +206,7 @@ class Projection(utils.SaveLoad):
                 u, s = stochastic_svd(
                     docs, k, chunksize=sys.maxsize,
                     num_terms=m, power_iters=self.power_iters,
-                    extra_dims=self.extra_dims, dtype=dtype)
+                    extra_dims=self.extra_dims, dtype=dtype, random_seed=self.random_seed)
             else:
                 try:
                     import sparsesvd
@@ -222,7 +234,10 @@ class Projection(utils.SaveLoad):
             An empty copy (without corpus) of the current projection.
 
         """
-        return Projection(self.m, self.k, power_iters=self.power_iters, extra_dims=self.extra_dims)
+        return Projection(
+            self.m, self.k, power_iters=self.power_iters,
+            extra_dims=self.extra_dims, random_seed=self.random_seed,
+        )
 
     def merge(self, other, decay=1.0):
         """Merge current :class:`~gensim.models.lsimodel.Projection` instance with another.
@@ -312,7 +327,7 @@ class Projection(utils.SaveLoad):
 
         # make each column of U start with a non-negative number (to force canonical decomposition)
         if self.u.shape[0] > 0:
-            for i in xrange(self.u.shape[1]):
+            for i in range(self.u.shape[1]):
                 if self.u[0, i] < 0.0:
                     self.u[:, i] *= -1.0
 
@@ -322,7 +337,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
     <https://en.wikipedia.org/wiki/Latent_semantic_analysis#Latent_semantic_indexing>`_.
 
     The decomposition algorithm is described in `"Fast and Faster: A Comparison of Two Streamed
-    Matrix Decomposition Algorithms" <https://nlp.fi.muni.cz/~xrehurek/nips/rehurek_nips.pdf>`_.
+    Matrix Decomposition Algorithms" <https://arxiv.org/pdf/1102.5597.pdf>`_.
 
     Notes
     -----
@@ -333,33 +348,35 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
     See Also
     --------
     `FAQ about LSI matrices
-    <https://github.com/piskvorky/gensim/wiki/Recipes-&-FAQ#q4-how-do-you-output-the-u-s-vt-matrices-of-lsi>`_.
+    <https://github.com/RaRe-Technologies/gensim/wiki/Recipes-&-FAQ#q4-how-do-you-output-the-u-s-vt-matrices-of-lsi>`_.
 
     Examples
     --------
-    >>> from gensim.test.utils import common_corpus, common_dictionary, get_tmpfile
-    >>> from gensim.models import LsiModel
-    >>>
-    >>> model = LsiModel(common_corpus[:3], id2word=common_dictionary)  # train model
-    >>> vector = model[common_corpus[4]]  # apply model to BoW document
-    >>> model.add_documents(common_corpus[4:])  # update model with new documents
-    >>> tmp_fname = get_tmpfile("lsi.model")
-    >>> model.save(tmp_fname)  # save model
-    >>> loaded_model = LsiModel.load(tmp_fname)  # load model
+    .. sourcecode:: pycon
+
+        >>> from gensim.test.utils import common_corpus, common_dictionary, get_tmpfile
+        >>> from gensim.models import LsiModel
+        >>>
+        >>> model = LsiModel(common_corpus[:3], id2word=common_dictionary)  # train model
+        >>> vector = model[common_corpus[4]]  # apply model to BoW document
+        >>> model.add_documents(common_corpus[4:])  # update model with new documents
+        >>> tmp_fname = get_tmpfile("lsi.model")
+        >>> model.save(tmp_fname)  # save model
+        >>> loaded_model = LsiModel.load(tmp_fname)  # load model
 
     """
 
-    def __init__(self, corpus=None, num_topics=200, id2word=None, chunksize=20000,
-                 decay=1.0, distributed=False, onepass=True,
-                 power_iters=P2_EXTRA_ITERS, extra_samples=P2_EXTRA_DIMS, dtype=np.float64):
-        """Construct an `LsiModel` object.
-
-        Either `corpus` or `id2word` must be supplied in order to train the model.
+    def __init__(
+            self, corpus=None, num_topics=200, id2word=None, chunksize=20000,
+            decay=1.0, distributed=False, onepass=True, power_iters=P2_EXTRA_ITERS,
+            extra_samples=P2_EXTRA_DIMS, dtype=np.float64, random_seed=None,
+    ):
+        """Build an LSI model.
 
         Parameters
         ----------
         corpus : {iterable of list of (int, float), scipy.sparse.csc}, optional
-            Stream of document vectors or sparse matrix of shape (`num_terms`, `num_documents`).
+            Stream of document vectors or a sparse matrix of shape (`num_documents`, `num_terms`).
         num_topics : int, optional
             Number of requested factors (latent dimensions)
         id2word : dict of {int: str}, optional
@@ -380,6 +397,9 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
             Extra samples to be used besides the rank `k`. Can improve accuracy.
         dtype : type, optional
             Enforces a type for elements of the decomposed matrix.
+        random_seed: {None, int}, optional
+            Random seed used to initialize the pseudo-random number generator,
+            a local instance of numpy.random.RandomState instance.
 
         """
         self.id2word = id2word
@@ -393,6 +413,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
         self.onepass = onepass
         self.extra_samples, self.power_iters = extra_samples, power_iters
         self.dtype = dtype
+        self.random_seed = random_seed
 
         if corpus is None and self.id2word is None:
             raise ValueError(
@@ -408,7 +429,8 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
 
         self.docs_processed = 0
         self.projection = Projection(
-            self.num_terms, self.num_topics, power_iters=self.power_iters, extra_dims=self.extra_samples, dtype=dtype
+            self.num_terms, self.num_topics, power_iters=self.power_iters,
+            extra_dims=self.extra_samples, dtype=dtype, random_seed=self.random_seed
         )
 
         self.numworkers = 1
@@ -438,7 +460,12 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
                 raise RuntimeError("failed to initialize distributed LSI (%s)" % err)
 
         if corpus is not None:
+            start = time.time()
             self.add_documents(corpus)
+            self.add_lifecycle_event(
+                "created",
+                msg=f"trained {self} in {time.time() - start:.2f}s",
+            )
 
     def add_documents(self, corpus, chunksize=None, decay=None):
         """Update model with new `corpus`.
@@ -466,15 +493,20 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
             chunksize = self.chunksize
         if decay is None:
             decay = self.decay
-
+        if is_empty(corpus):
+            logger.warning('LsiModel.add_documents() called but no documents provided, is this intended?')
         if not scipy.sparse.issparse(corpus):
             if not self.onepass:
                 # we are allowed multiple passes over the input => use a faster, randomized two-pass algo
-                update = Projection(self.num_terms, self.num_topics, None, dtype=self.dtype)
+                update = Projection(
+                    self.num_terms, self.num_topics, None,
+                    dtype=self.dtype, random_seed=self.random_seed,
+                )
                 update.u, update.s = stochastic_svd(
                     corpus, self.num_topics,
                     num_terms=self.num_terms, chunksize=chunksize,
-                    extra_dims=self.extra_samples, power_iters=self.power_iters, dtype=self.dtype
+                    extra_dims=self.extra_samples, power_iters=self.power_iters, dtype=self.dtype,
+                    random_seed=self.random_seed,
                 )
                 self.projection.merge(update, decay=decay)
                 self.docs_processed += len(corpus) if hasattr(corpus, '__len__') else 0
@@ -491,7 +523,9 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
                     # definitely avoid materializing it as a dense matrix!
                     logger.debug("converting corpus to csc format")
                     job = matutils.corpus2csc(
-                        chunk, num_docs=len(chunk), num_terms=self.num_terms, num_nnz=nnz, dtype=self.dtype)
+                        chunk, num_docs=len(chunk), num_terms=self.num_terms,
+                        num_nnz=nnz, dtype=self.dtype,
+                    )
                     del chunk
                     doc_no += job.shape[1]
                     if self.dispatcher:
@@ -505,7 +539,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
                         # serial version, there is only one "worker" (myself) => process the job directly
                         update = Projection(
                             self.num_terms, self.num_topics, job, extra_dims=self.extra_samples,
-                            power_iters=self.power_iters, dtype=self.dtype
+                            power_iters=self.power_iters, dtype=self.dtype, random_seed=self.random_seed,
                         )
                         del job
                         self.projection.merge(update, decay=decay)
@@ -522,7 +556,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
             assert not self.dispatcher, "must be in serial mode to receive jobs"
             update = Projection(
                 self.num_terms, self.num_topics, corpus.tocsc(), extra_dims=self.extra_samples,
-                power_iters=self.power_iters, dtype=self.dtype
+                power_iters=self.power_iters, dtype=self.dtype,
             )
             self.projection.merge(update, decay=decay)
             logger.info("processed sparse job of %i documents", corpus.shape[1])
@@ -537,8 +571,8 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
             A human readable string of the current objects parameters.
 
         """
-        return "LsiModel(num_terms=%s, num_topics=%s, decay=%s, chunksize=%s)" % (
-            self.num_terms, self.num_topics, self.decay, self.chunksize
+        return "%s<num_terms=%s, num_topics=%s, decay=%s, chunksize=%s>" % (
+            self.__class__.__name__, self.num_terms, self.num_topics, self.decay, self.chunksize
         )
 
     def __getitem__(self, bow, scaled=False, chunksize=512):
@@ -561,7 +595,8 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
             Latent representation of corpus in BoW format if `bow` is corpus.
 
         """
-        assert self.projection.u is not None, "decomposition not initialized yet"
+        if self.projection.u is None:
+            raise ValueError('No training data provided - LSI model not initialized yet')
 
         # if the input vector is in fact a corpus, return a transformed corpus as a result
         is_corpus, bow = utils.is_corpus(bow)
@@ -662,7 +697,9 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
         c = np.asarray(self.projection.u.T[topicno, :]).flatten()
         norm = np.sqrt(np.sum(np.dot(c, c)))
         most = matutils.argsort(np.abs(c), topn, reverse=True)
-        return [(self.id2word[val], 1.0 * c[val] / norm) for val in most]
+
+        # Output only (word, score) pairs for `val`s that are within `self.id2word`.  See #3090 for details.
+        return [(self.id2word[val], 1.0 * c[val] / norm) for val in most if val in self.id2word]
 
     def show_topics(self, num_topics=-1, num_words=10, log=False, formatted=True):
         """Get the most significant topics.
@@ -689,7 +726,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
         shown = []
         if num_topics < 0:
             num_topics = self.num_topics
-        for i in xrange(min(num_topics, self.num_topics)):
+        for i in range(min(num_topics, self.num_topics)):
             if i < len(self.projection.s):
                 if formatted:
                     topic = self.print_topic(i, topn=num_words)
@@ -721,7 +758,7 @@ class LsiModel(interfaces.TransformationABC, basemodel.BaseTopicModel):
         print_debug(
             self.id2word, self.projection.u, self.projection.s,
             range(min(num_topics, len(self.projection.u.T))),
-            num_words=num_words
+            num_words=num_words,
         )
 
     def save(self, fname, *args, **kwargs):
@@ -829,7 +866,7 @@ def print_debug(id2token, u, s, topics, num_words=10, num_neg=None):
             result.setdefault(topic, []).append((udiff[topic], uvecno))
 
     logger.debug("printing %i+%i salient words", num_words, num_neg)
-    for topic in sorted(iterkeys(result)):
+    for topic in sorted(result.keys()):
         weights = sorted(result[topic], key=lambda x: -abs(x[0]))
         _, most = weights[0]
         if u[most, topic] < 0.0:  # the most significant word has a negative sign => flip sign of u[most]
@@ -854,8 +891,10 @@ def print_debug(id2token, u, s, topics, num_words=10, num_neg=None):
         logger.info('topic #%s(%.3f): %s, ..., %s', topic, s[topic], ', '.join(pos), ', '.join(neg))
 
 
-def stochastic_svd(corpus, rank, num_terms, chunksize=20000, extra_dims=None,
-                   power_iters=0, dtype=np.float64, eps=1e-6):
+def stochastic_svd(
+        corpus, rank, num_terms, chunksize=20000, extra_dims=None,
+        power_iters=0, dtype=np.float64, eps=1e-6, random_seed=None,
+    ):
     """Run truncated Singular Value Decomposition (SVD) on a sparse input.
 
     Parameters
@@ -878,6 +917,10 @@ def stochastic_svd(corpus, rank, num_terms, chunksize=20000, extra_dims=None,
         Enforces a type for elements of the decomposed matrix.
     eps: float, optional
         Percentage of the spectrum's energy to be discarded.
+    random_seed: {None, int}, optional
+        Random seed used to initialize the pseudo-random number generator,
+         a local instance of numpy.random.RandomState instance.
+
 
     Notes
     -----
@@ -914,13 +957,17 @@ def stochastic_svd(corpus, rank, num_terms, chunksize=20000, extra_dims=None,
     # and more memory friendly than processing all documents at once)
     y = np.zeros(dtype=dtype, shape=(num_terms, samples))
     logger.info("1st phase: constructing %s action matrix", str(y.shape))
+    random_state = np.random.RandomState(random_seed)
 
     if scipy.sparse.issparse(corpus):
         m, n = corpus.shape
-        assert num_terms == m, "mismatch in number of features: %i in sparse matrix vs. %i parameter" % (m, num_terms)
-        o = np.random.normal(0.0, 1.0, (n, samples)).astype(y.dtype)  # draw a random gaussian matrix
-        sparsetools.csc_matvecs(m, n, samples, corpus.indptr, corpus.indices,
-                                corpus.data, o.ravel(), y.ravel())  # y = corpus * o
+        assert num_terms == m, f"mismatch in number of features: {m} in sparse matrix vs. {num_terms} parameter"
+        o = random_state.normal(0.0, 1.0, (n, samples)).astype(y.dtype)  # draw a random gaussian matrix
+        csc_matvecs(
+            m, n, samples, corpus.indptr, corpus.indices,
+            corpus.data, o.ravel(), y.ravel(),
+        )  # y = corpus * o
+
         del o
 
         # unlike np, scipy.sparse `astype()` copies everything, even if there is no change to dtype!
@@ -933,7 +980,7 @@ def stochastic_svd(corpus, rank, num_terms, chunksize=20000, extra_dims=None,
         q, _ = matutils.qr_destroy(y)  # orthonormalize the range
 
         logger.debug("running %i power iterations", power_iters)
-        for _ in xrange(power_iters):
+        for _ in range(power_iters):
             q = corpus.T * q
             q = [corpus * q]
             q, _ = matutils.qr_destroy(q)  # orthonormalize the range after each power iteration step
@@ -950,16 +997,16 @@ def stochastic_svd(corpus, rank, num_terms, chunksize=20000, extra_dims=None,
             assert n <= chunksize  # the very last chunk of A is allowed to be smaller in size
             num_docs += n
             logger.debug("multiplying chunk * gauss")
-            o = np.random.normal(0.0, 1.0, (n, samples)).astype(dtype)  # draw a random gaussian matrix
-            sparsetools.csc_matvecs(
+            o = random_state.normal(0.0, 1.0, (n, samples), ).astype(dtype)  # draw a random gaussian matrix
+            csc_matvecs(
                 m, n, samples, chunk.indptr, chunk.indices,  # y = y + chunk * o
-                chunk.data, o.ravel(), y.ravel()
+                chunk.data, o.ravel(), y.ravel(),
             )
             del chunk, o
         y = [y]
         q, _ = matutils.qr_destroy(y)  # orthonormalize the range
 
-        for power_iter in xrange(power_iters):
+        for power_iter in range(power_iters):
             logger.info("running power iteration #%i", power_iter + 1)
             yold = q.copy()
             q[:] = 0.0
