@@ -3,7 +3,7 @@
 #
 # Author: Jayant Jain <jayantjain1992@gmail.com>
 # Copyright (C) 2017 Radim Rehurek <me@radimrehurek.com>
-# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html
+# Licensed under the GNU LGPL v2.1 - https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 
 """
 Automated tests for checking the poincare module from the models package.
@@ -13,10 +13,7 @@ import logging
 import os
 import tempfile
 import unittest
-try:
-    from mock import Mock
-except ImportError:
-    from unittest.mock import Mock
+from unittest.mock import Mock
 
 import numpy as np
 try:
@@ -32,7 +29,7 @@ from gensim.test.utils import datapath
 logger = logging.getLogger(__name__)
 
 
-def testfile():
+def get_testfile():
     # temporary data will be stored to this file
     return os.path.join(tempfile.gettempdir(), 'gensim_word2vec.tst')
 
@@ -57,48 +54,58 @@ class TestPoincareModel(unittest.TestCase):
         self.data_large = PoincareRelations(datapath('poincare_hypernyms_large.tsv'))
 
     def models_equal(self, model_1, model_2):
-        self.assertEqual(len(model_1.kv.vocab), len(model_2.kv.vocab))
-        self.assertEqual(set(model_1.kv.vocab.keys()), set(model_2.kv.vocab.keys()))
-        self.assertTrue(np.allclose(model_1.kv.syn0, model_2.kv.syn0))
+        self.assertEqual(len(model_1.kv), len(model_2.kv))
+        self.assertEqual(set(model_1.kv.index_to_key), set(model_2.kv.index_to_key))
+        self.assertTrue(np.allclose(model_1.kv.vectors, model_2.kv.vectors))
 
     def test_data_counts(self):
         """Tests whether data has been loaded correctly and completely."""
         model = PoincareModel(self.data)
         self.assertEqual(len(model.all_relations), 5)
-        self.assertEqual(len(model.node_relations[model.kv.vocab['kangaroo.n.01'].index]), 3)
-        self.assertEqual(len(model.kv.vocab), 7)
+        self.assertEqual(len(model.node_relations[model.kv.get_index('kangaroo.n.01')]), 3)
+        self.assertEqual(len(model.kv), 7)
         self.assertTrue('mammal.n.01' not in model.node_relations)
 
     def test_data_counts_with_bytes(self):
         """Tests whether input bytes data is loaded correctly and completely."""
         model = PoincareModel([(b'\x80\x01c', b'\x50\x71a'), (b'node.1', b'node.2')])
         self.assertEqual(len(model.all_relations), 2)
-        self.assertEqual(len(model.node_relations[model.kv.vocab[b'\x80\x01c'].index]), 1)
-        self.assertEqual(len(model.kv.vocab), 4)
+        self.assertEqual(len(model.node_relations[model.kv.get_index(b'\x80\x01c')]), 1)
+        self.assertEqual(len(model.kv), 4)
         self.assertTrue(b'\x50\x71a' not in model.node_relations)
 
     def test_persistence(self):
         """Tests whether the model is saved and loaded correctly."""
         model = PoincareModel(self.data, burn_in=0, negative=3)
         model.train(epochs=1)
-        model.save(testfile())
-        loaded = PoincareModel.load(testfile())
+        model.save(get_testfile())
+        loaded = PoincareModel.load(get_testfile())
         self.models_equal(model, loaded)
 
     def test_persistence_separate_file(self):
         """Tests whether the model is saved and loaded correctly when the arrays are stored separately."""
         model = PoincareModel(self.data, burn_in=0, negative=3)
         model.train(epochs=1)
-        model.save(testfile(), sep_limit=1)
-        loaded = PoincareModel.load(testfile())
+        model.save(get_testfile(), sep_limit=1)
+        loaded = PoincareModel.load(get_testfile())
         self.models_equal(model, loaded)
 
+    def test_online_learning(self):
+        """Tests whether additional input data is loaded correctly and completely."""
+        model = PoincareModel(self.data, burn_in=0, negative=3)
+        self.assertEqual(len(model.kv), 7)
+        self.assertEqual(model.kv.get_vecattr('kangaroo.n.01', 'count'), 3)
+        self.assertEqual(model.kv.get_vecattr('cat.n.01', 'count'), 1)
+        model.build_vocab([('kangaroo.n.01', 'cat.n.01')], update=True)  # update vocab
+        self.assertEqual(model.kv.get_vecattr('kangaroo.n.01', 'count'), 4)
+        self.assertEqual(model.kv.get_vecattr('cat.n.01', 'count'), 2)
+
     def test_train_after_load(self):
         """Tests whether the model can be trained correctly after loading from disk."""
         model = PoincareModel(self.data, burn_in=0, negative=3)
         model.train(epochs=1)
-        model.save(testfile())
-        loaded = PoincareModel.load(testfile())
+        model.save(get_testfile())
+        loaded = PoincareModel.load(get_testfile())
         model.train(epochs=1)
         loaded.train(epochs=1)
         self.models_equal(model, loaded)
@@ -106,17 +113,17 @@ class TestPoincareModel(unittest.TestCase):
     def test_persistence_old_model(self):
         """Tests whether model from older gensim version is loaded correctly."""
         loaded = PoincareModel.load(datapath('poincare_test_3.4.0'))
-        self.assertEqual(loaded.kv.syn0.shape, (239, 2))
-        self.assertEqual(len(loaded.kv.vocab), 239)
+        self.assertEqual(loaded.kv.vectors.shape, (239, 2))
+        self.assertEqual(len(loaded.kv), 239)
         self.assertEqual(loaded.size, 2)
         self.assertEqual(len(loaded.all_relations), 200)
 
     def test_train_old_model_after_load(self):
         """Tests whether loaded model from older gensim version can be trained correctly."""
         loaded = PoincareModel.load(datapath('poincare_test_3.4.0'))
-        old_vectors = np.copy(loaded.kv.syn0)
+        old_vectors = np.copy(loaded.kv.vectors)
         loaded.train(epochs=2)
-        self.assertFalse(np.allclose(old_vectors, loaded.kv.syn0))
+        self.assertFalse(np.allclose(old_vectors, loaded.kv.vectors))
 
     def test_invalid_data_raises_error(self):
         """Tests that error is raised on invalid input data."""
@@ -130,34 +137,34 @@ class TestPoincareModel(unittest.TestCase):
     def test_vector_shape(self):
         """Tests whether vectors are initialized with the correct size."""
         model = PoincareModel(self.data, size=20)
-        self.assertEqual(model.kv.syn0.shape, (7, 20))
+        self.assertEqual(model.kv.vectors.shape, (7, 20))
 
     def test_vector_dtype(self):
         """Tests whether vectors have the correct dtype before and after training."""
         model = PoincareModel(self.data_large, dtype=np.float32, burn_in=0, negative=3)
-        self.assertEqual(model.kv.syn0.dtype, np.float32)
+        self.assertEqual(model.kv.vectors.dtype, np.float32)
         model.train(epochs=1)
-        self.assertEqual(model.kv.syn0.dtype, np.float32)
+        self.assertEqual(model.kv.vectors.dtype, np.float32)
 
     def test_training(self):
         """Tests that vectors are different before and after training."""
         model = PoincareModel(self.data_large, burn_in=0, negative=3)
-        old_vectors = np.copy(model.kv.syn0)
+        old_vectors = np.copy(model.kv.vectors)
         model.train(epochs=2)
-        self.assertFalse(np.allclose(old_vectors, model.kv.syn0))
+        self.assertFalse(np.allclose(old_vectors, model.kv.vectors))
 
     def test_training_multiple(self):
         """Tests that calling train multiple times results in different vectors."""
         model = PoincareModel(self.data_large, burn_in=0, negative=3)
         model.train(epochs=2)
-        old_vectors = np.copy(model.kv.syn0)
+        old_vectors = np.copy(model.kv.vectors)
 
         model.train(epochs=1)
-        self.assertFalse(np.allclose(old_vectors, model.kv.syn0))
+        self.assertFalse(np.allclose(old_vectors, model.kv.vectors))
 
-        old_vectors = np.copy(model.kv.syn0)
+        old_vectors = np.copy(model.kv.vectors)
         model.train(epochs=0)
-        self.assertTrue(np.allclose(old_vectors, model.kv.syn0))
+        self.assertTrue(np.allclose(old_vectors, model.kv.vectors))
 
     def test_gradients_check(self):
         """Tests that the model is trained successfully with gradients check enabled."""
@@ -182,22 +189,22 @@ class TestPoincareModel(unittest.TestCase):
 
         model_2 = PoincareModel(self.data_large, seed=1, negative=3, burn_in=1)
         model_2.train(epochs=2)
-        self.assertTrue(np.allclose(model_1.kv.syn0, model_2.kv.syn0))
+        self.assertTrue(np.allclose(model_1.kv.vectors, model_2.kv.vectors))
 
     def test_burn_in(self):
         """Tests that vectors are different after burn-in."""
         model = PoincareModel(self.data, burn_in=1, negative=3)
-        original_vectors = np.copy(model.kv.syn0)
+        original_vectors = np.copy(model.kv.vectors)
         model.train(epochs=0)
-        self.assertFalse(np.allclose(model.kv.syn0, original_vectors))
+        self.assertFalse(np.allclose(model.kv.vectors, original_vectors))
 
     def test_burn_in_only_done_once(self):
         """Tests that burn-in does not happen when train is called a second time."""
         model = PoincareModel(self.data, negative=3, burn_in=1)
         model.train(epochs=0)
-        original_vectors = np.copy(model.kv.syn0)
+        original_vectors = np.copy(model.kv.vectors)
         model.train(epochs=0)
-        self.assertTrue(np.allclose(model.kv.syn0, original_vectors))
+        self.assertTrue(np.allclose(model.kv.vectors, original_vectors))
 
     def test_negatives(self):
         """Tests that correct number of negatives are sampled."""
@@ -231,7 +238,7 @@ class TestPoincareModel(unittest.TestCase):
     @classmethod
     def tearDownClass(cls):
         try:
-            os.unlink(testfile())
+            os.unlink(get_testfile())
         except OSError:
             pass
 
@@ -258,7 +265,7 @@ class TestPoincareKeyedVectors(unittest.TestCase):
         self.assertEqual(len(self.vectors.most_similar('dog.n.01', topn=10)), 10)
 
         predicted = self.vectors.most_similar('dog.n.01', topn=None)
-        self.assertEqual(len(predicted), len(self.vectors.vocab) - 1)
+        self.assertEqual(len(predicted), len(self.vectors) - 1)
         self.assertEqual(predicted[-1][0], 'gallant_fox.n.01')
 
     def test_most_similar_raises_keyerror(self):
@@ -268,7 +275,7 @@ class TestPoincareKeyedVectors(unittest.TestCase):
 
     def test_most_similar_restrict_vocab(self):
         """Test most_similar returns handles restrict_vocab correctly."""
-        expected = set(self.vectors.index2word[:5])
+        expected = set(self.vectors.index_to_key[:5])
         predicted = set(result[0] for result in self.vectors.most_similar('dog.n.01', topn=5, restrict_vocab=5))
         self.assertEqual(expected, predicted)
 
@@ -301,7 +308,7 @@ class TestPoincareKeyedVectors(unittest.TestCase):
         self.assertTrue(np.allclose(distances, [4.5278745, 0]))
 
         distances = self.vectors.distances('dog.n.01')
-        self.assertEqual(len(distances), len(self.vectors.vocab))
+        self.assertEqual(len(distances), len(self.vectors))
         self.assertTrue(np.allclose(distances[-1], 10.04756))
 
     def test_distances_with_vector_input(self):
@@ -311,7 +318,7 @@ class TestPoincareKeyedVectors(unittest.TestCase):
         self.assertTrue(np.allclose(distances, [4.5278745, 0]))
 
         distances = self.vectors.distances(input_vector)
-        self.assertEqual(len(distances), len(self.vectors.vocab))
+        self.assertEqual(len(distances), len(self.vectors))
         self.assertTrue(np.allclose(distances[-1], 10.04756))
 
     def test_poincare_distances_batch(self):
@@ -373,11 +380,11 @@ class TestPoincareKeyedVectors(unittest.TestCase):
         self.assertTrue(np.allclose(self.vectors.difference_in_hierarchy('mammal.n.01', 'dog.n.01'), 0.9384287))
         self.assertTrue(np.allclose(self.vectors.difference_in_hierarchy('dog.n.01', 'mammal.n.01'), -0.9384287))
 
-    def test_words_closer_than(self):
-        """Test words_closer_than returns expected value for distinct and identical nodes."""
-        self.assertEqual(self.vectors.words_closer_than('dog.n.01', 'dog.n.01'), [])
+    def test_closer_than(self):
+        """Test closer_than returns expected value for distinct and identical nodes."""
+        self.assertEqual(self.vectors.closer_than('dog.n.01', 'dog.n.01'), [])
         expected = set(['canine.n.02', 'hunting_dog.n.01'])
-        self.assertEqual(set(self.vectors.words_closer_than('dog.n.01', 'carnivore.n.01')), expected)
+        self.assertEqual(set(self.vectors.closer_than('dog.n.01', 'carnivore.n.01')), expected)
 
     def test_rank(self):
         """Test rank returns expected value for distinct and identical nodes."""
